//////////////////////////////////////////////////////////////////////////////
// CUDA exercise 1: Memory Transfer
//////////////////////////////////////////////////////////////////////////////

// includes
#include <stdio.h>

void checkCUDAError(const char *msg)
{
	cudaError_t err = cudaGetLastError();
	if (cudaSuccess != err)
	{
		printf("Cuda error: %s: %s.\n", msg, cudaGetErrorString(err));
		exit(-1);
	}
}

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	////////////////////////  Declarations ////////////////////////

	// pointers for host memory
	int *h_a_cpu; //pointer for CPU results
	int *h_a_gpu; //pointers for GPU results on host

	// pointer for device memory
	int *d_a;//pointers for GPU results on device

	// number of elements and memory size
	int nElem = 1024;
	size_t memSize = nElem * sizeof(int);

	////////////////////////  Allocate Memory   ////////////////////////

	//Allocate host memory
	h_a_cpu = (int *) malloc(memSize);
	h_a_gpu = (int *) malloc(memSize);

	//Allocate device memory
	cudaMalloc((void **) &d_a, memSize);
	// Check for any CUDA errors
	checkCUDAError("cudaMalloc");

	////////////////////////  Initialize host memory  ////////////////////////
	for (int i = 0; i < nElem; i = i + 1)
	{
		h_a_cpu[i] = i;
	}

	////////////////////////  Copy Data From Host To Device   ////////////////////////
	cudaMemcpy(d_a, h_a_cpu, memSize, cudaMemcpyHostToDevice);
	// Check for any CUDA errors
	checkCUDAError("cudaMemcpyHostToDevice");

	////////////////////////  Copy Data From Device To Host   ////////////////////////
	cudaMemcpy(h_a_gpu, d_a, memSize, cudaMemcpyDeviceToHost);
	// Check for any CUDA errors
	checkCUDAError("cudaMemcpyDeviceToHost");

	////////////////////////  Compare the results             ////////////////////////
	for (int i = 0; i < nElem; i++)
	{
		if (h_a_gpu[i] != h_a_cpu[i])
		{
			printf("Error : i= %d h_a_gpu[i]= %d h_a_cpu[i]= %d \n", i,
					h_a_gpu[i], h_a_cpu[i]);
			printf("\nPress ENTER to exit...\n");
			getchar();
			return 1;
		}
	}

	////////////////////////  Free Memory      ////////////////////////
	// free device memory
	cudaFree(d_a);

	// free host memory
	free(h_a_cpu);
	free(h_a_gpu);

	////////////////////////  End    ////////////////////////
	printf("Correct!\n");
	printf("\nPress ENTER to exit...\n");
	getchar();
	return 0;
}

