#include <stdio.h>

void  initVector(double *A, int n, double c) {
  int i;
  for (i=0; i<n; i++)
      A[i] = c;
}

/* insert the appropriate qualifier for the kernel */
__global__ void gpuIncVect(double *A, double b, int N) 
{
/* define index on a monodimensional grid */
  int idx =threadIdx.x+blockIdx.x*blockDim.x;
/* check for active threads and add the scalar b to vector A */
  if(idx>=N) return;
  A[idx]=A[idx]+b;
}

int main(int argc, char *argv[]) {
  const int N = 100000;
  double A[N];
  const double b = 2.0;

  double *A_dev; 
// allocate the device memory for matrix A
  cudaMalloc((void **)&A_dev,sizeof(double)*N);

  initVector((double *) A, N, 1.0);

// copy vector A from Host to Device
  cudaMemcpy(A_dev,A,N*sizeof(double),cudaMemcpyHostToDevice);
  dim3 mblockDim(128);
  dim3 mgridDim(N/mblockDim.x + 1);

// specify the execution configuration of the CUDA kernel
   gpuIncVect<<<mgridDim,mblockDim>>>(A_dev,b,N);

// copy vector A from Device to Host
   cudaMemcpy(A,A_dev,N*sizeof(double),cudaMemcpyDeviceToHost);

  printf("%f %f %f\n", A[0], A[1], A[1]);
  cudaFree(A_dev);

  return 0;
}

