#include <stdio.h>

void  initMatrix(double *A, int n, int m, double c) {
  int i,j;
  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
      A[i*m+j] = c;
}

__global__ void gpuMatAdd(double *A, double *B,
  double *C, int N, int M) 
{
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;

  if (i >= M || j >= N) return;

  int index = j*M + i; 
  C[index] = A[index] + B[index];
}

int main(int argc, char *argv[]) {
  const int N = 1000, M = 1000;
  double A[N][M], B[N][M], C[N][M];

  double *A_dev, *B_dev, *C_dev;
  cudaMalloc((void **) &A_dev, M*N*sizeof(double));
  cudaMalloc((void **) &B_dev, M*N*sizeof(double));
  cudaMalloc((void **) &C_dev, M*N*sizeof(double));

  initMatrix((double *) A, N, M, 1.0);
  initMatrix((double *) B, N, M, 2.0);
  initMatrix((double *) C, N, M, 0.0);

  cudaMemcpy(A_dev, A, M*N*sizeof(double), cudaMemcpyHostToDevice);
  cudaMemcpy(B_dev, B, M*N*sizeof(double), cudaMemcpyHostToDevice);

  dim3 mblockDim(32,16);
  dim3 mgridDim(M/mblockDim.x + 1, N/mblockDim.y + 1);

  gpuMatAdd<<<mgridDim, mblockDim>>>(A_dev, B_dev, C_dev, N, M);

  cudaMemcpy(C, C_dev, M*N*sizeof(double), cudaMemcpyDeviceToHost);

  printf("%f %f %f\n", C[0][0], C[1][1], C[1][3]);

  cudaFree(A_dev);
  cudaFree(B_dev);
  cudaFree(C_dev);

  return 0;
}

