#include <stdio.h>

void  initMatrix(double *A, int n, int m, double c) {
  int i,j;
  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
      A[i*m+j] = c+i-j;
}

void  printMatrix(double *A, int n, int m) {
  int i,j;
    for (j=0; j<m; j++)
	for(i=0;i<n;i++)
      printf("%d  %d  %lf \n",i,j,A[i*m+j]);
}

__global__ void gpu_stencil2d(double *A,
  double *C, int N, int M, int N_halo) 
{
  int i_h, j_h, index_shifted;
//   Definizione indici di matrice calcolati dal thread corrente
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;

//   Condizione per lasciare solo i thread attivi (lontani dal bordo)
  if((i < N_halo) || (j < N_halo) || (j >= N-N_halo) || (i >= M-N_halo)) return;

//   Ciclo di calcolo 
  int index = j*M + i;
  C[index] = 0. ; 
  for (i_h=-N_halo; i_h<=N_halo; i_h++)
      for (j_h=-N_halo; j_h<=N_halo; j_h++)  
          {
             index_shifted = (j+j_h)*M+(i+i_h);
             C[index] = C[index] + A[index_shifted] ;
          }
   C[index] = C[index] / ((2*N_halo+1)*(2*N_halo+1)) ;
}

int main(int argc, char *argv[]) {
  const int N = 5, M = 5, N_halo = 2;
  double A[N][M], C[N][M];

  double *A_dev, *C_dev;
  cudaMalloc((void **) &A_dev, M*N*sizeof(double));
  cudaMalloc((void **) &C_dev, M*N*sizeof(double));

  initMatrix((double *) A, N, M, 1.0);
  initMatrix((double *) C, N, M, 7.0);

// Copia da host a device cudaMemcpy A e C
  cudaMemcpy(A_dev, A, M*N*sizeof(double), cudaMemcpyHostToDevice);
  cudaMemcpy(C_dev, C, M*N*sizeof(double), cudaMemcpyHostToDevice);

// Definizione dimensioni blocco e griglia di thread
  dim3 mblockDim(32,16);
  dim3 mgridDim(M/mblockDim.x + 1, N/mblockDim.y + 1);

  gpu_stencil2d<<<mgridDim, mblockDim>>>(A_dev, C_dev, N, M, N_halo);

// Copia da device a host cudaMemcpy C
  cudaMemcpy(C, C_dev, M*N*sizeof(double), cudaMemcpyDeviceToHost);

//  printf("%f %f %f\n", C[2][2], C[1][1], C[1][2]);
  printMatrix((double *) C, N, M);

  cudaFree(A_dev);
  cudaFree(C_dev);

  return 0;
}
