#include <stdio.h>
#include <stdlib.h>


#define N (1024)
#define THREADS_PER_BLOCK 64

//insert the appropriate qualifier for the kernel 
//just do product and compute reduce on host
__global__ void gpuProduct(float *A_dev ,float *B_dev, float *C_dev) 
{
//define index on a monodimensional grid 
  int idx = blockIdx.x*blockDim.x+threadIdx.x;
// check for active threads and add the scalar b to vector A 
	if(idx > N) return;
	C_dev[idx]=A_dev[idx]*B_dev[idx];  

}


//Compute product and reduce
__global__ void gpuInner(float *A ,float *B, float *C)
{
	
//define index on a monodimensional grid 
  int idx = blockIdx.x*blockDim.x+threadIdx.x;
  
// check for active threads 
      if(idx> N) return;
      __shared__ float temp[THREADS_PER_BLOCK];
      temp[threadIdx.x]=A[idx]*B[idx];  
      __syncthreads();
    
      if(threadIdx.x==0){
    	  float sum=0.0;
	  //Compute partial sum
    	  for(int i=0;i<THREADS_PER_BLOCK;i++) sum+=temp[i];
    	  atomicAdd(C,sum);
	  //Why not??
    	  //*C+=sum;
	} 
	
}


void  initVector(float *A, float c) {
  for (int i=0; i<N; i++) A[i] = c;
}

void cpuInner(float *A,float *B, float *C){
  float sum=0.0;
  for(int i=0;i<N;i++) sum+=A[i]*B[i];
  *C=sum;
}


int main(int argc, char *argv[]) {
  float *A,*B,C;
  float *Cv;
  float *A_dev, *B_dev, *C_dev,*Cv_dev; 
  
  
  //allocate memory on host
  A  = (float*)malloc( N*sizeof(float) );
  B  = (float*)malloc( N*sizeof(float) );
  Cv = (float*)malloc( N*sizeof(float) );
  
  
  initVector((float *) A,  1.0);
  initVector((float *) B,  2.0);
  initVector((float *) Cv, 0.0);
  
  
  C=0.0;
  //allocate memory on  device 
  cudaMalloc((void**)&A_dev,N*sizeof(float));	
  cudaMalloc((void**)&B_dev,N*sizeof(float));
  cudaMalloc((void**)&Cv_dev,N*sizeof(float));
  cudaMalloc((void**)&C_dev,sizeof(float));
  
  //copy vector  from Host to Device
  cudaMemcpy(A_dev,A,N*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(B_dev,B,N*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(Cv_dev,Cv,N*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(C_dev,&C,sizeof(C),cudaMemcpyHostToDevice);
 
  dim3 mblockDim(THREADS_PER_BLOCK);
  dim3 mgridDim(N/THREADS_PER_BLOCK);

  //specify the execution configuration of the CUDA kernel
  printf("Method 1:\n");
  gpuProduct<<<mgridDim,mblockDim>>>(A_dev,B_dev,Cv_dev);
  cudaMemcpy(Cv,Cv_dev,N*sizeof(float),cudaMemcpyDeviceToHost); 	
  float sum=0.0;
  for(int i=0;i<N;i++) sum+=Cv[i];
  printf("Inner Product on GPU =%f\n",sum);

  printf("Method 2:\n");
  gpuInner<<<mgridDim,mblockDim>>>(A_dev,B_dev,C_dev);
  
  //copy vector from Device to Host
  cudaMemcpy(&C,C_dev,sizeof(float),cudaMemcpyDeviceToHost);
  printf("Inner Product on GPU :%f\n", C);
  
  //Computing on Host
  cpuInner(A,B,&C);
  printf("Inner Product on CPU: %f\n", C);

  
  //Free Memory
  cudaFree(A_dev);
  cudaFree(B_dev);
  cudaFree(C_dev);
  free(A);
  free(B);
  free(Cv);
  return 0;
}

