#include<stdio.h>
#include<stdlib.h>
#include<math.h>


#define N 20480
#define THREADS_PER_BLOCK 512

//CPU Approximation 
float piEstimateCPU(float *x, float *y)
{
    int s=0;
    for( int i = 0; i < N; i++)
    {
    	float xx = x[i];
		float yy = y[i];
		float norm = xx*xx + yy*yy;
		if (norm < 1.0f)	
			s++;		
    }
   return s*4.0/N;

}



__global__ void piEstimateGPU(float *x_dev,float *y_dev, int *count ){
	float xx,yy, norm;
	int idx=blockIdx.x*blockDim.x+threadIdx.x;
	if(idx>= N) return;
	
	xx=x_dev[idx];
	yy=y_dev[idx];
	norm =xx*xx+yy*yy;
	if (norm <1.0f)  count[idx]=1;
	else count[idx]=0;
		
}



//Uniform Number Generator
float unifRand()
{
    return rand() /float(RAND_MAX);
}


int main(){
	
	float *x,*y;
	float *x_dev,*y_dev;
	float p_cpu, p_dev=0.0;
	
	int *count_dev, *count;
	


	//Allocate memory on Host
	x=(float*)malloc(N*sizeof(float));
        y=(float*)malloc(N*sizeof(float));
	count=(int*)malloc(N*sizeof(int));

	//Inizialize Host  
	for (int i=0;i<N;i++){
		x[i]=unifRand();
		y[i]=unifRand();	
		count[i]=0;
	}
	
	//Allocate memory on Device
	cudaMalloc((void**)&x_dev,N*sizeof(float));	
	cudaMalloc((void**)&y_dev,N*sizeof(float));
	cudaMalloc((void**)&count_dev,N*sizeof(int));
	
	
	
	//Transfer memory from  host to device
	cudaMemcpy(x_dev,x,N*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(y_dev,y,N*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(count_dev,count,N*sizeof(int),cudaMemcpyHostToDevice);
	

	
	p_cpu=piEstimateCPU(x,y);
	printf("CPU pi=%f\n", p_cpu);
	
	
	//Define grid dimension and block dimension
	dim3 mblockDim(THREADS_PER_BLOCK);
	dim3 mgridDim(N/THREADS_PER_BLOCK+1);
	//Call kernel
	piEstimateGPU<<<mgridDim,mblockDim>>>(x_dev,y_dev,count_dev);
	cudaMemcpy(count,count_dev,N*sizeof(int),cudaMemcpyDeviceToHost);
	
	
	//Reduce

	for(int i=0;i<N;i++) p_dev+=count[i];
	printf("GPU pi=%f\n", p_dev*4.0/N);

	
	return 0;
}
