#include<stdio.h>
#include<stdlib.h>
#include<math.h>
 
 
 
#define N 20480
#define THREADS_PER_BLOCK 512
 
//CPU Approximation
float piEstimateCPU(float *x, float *y)
{
    int s=0;
    for( int i = 0; i < N; i++)
    {
        float xx = x[i];
                float yy = y[i];
                float norm = xx*xx + yy*yy;
                if (norm < 1.0f)
                        s++;
    }
   return s*4.0/N;
 
}
 
 
__global__ void piEstimateGPU(float *x_dev,float *y_dev, float *c){
        float xx,yy, norm;
        int idx=blockIdx.x*blockDim.x+threadIdx.x;
        if(idx>=N) return;
        xx=x_dev[idx];
        yy=y_dev[idx];
        norm =xx*xx+yy*yy;
        if (norm <1.0f) {
        float pi=4.0/N;
        atomicAdd(c,pi);
        }
 
}
 
__global__ void piEstimateGPUInt(float *x_dev,float *y_dev, int *c){
        float xx,yy, norm;
        int idx=blockIdx.x*blockDim.x+threadIdx.x;
        if(idx>=N) return;
        xx=x_dev[idx];
        yy=y_dev[idx];
        norm =xx*xx+yy*yy;
        if (norm <1.0f) {
        atomicAdd(c,1);
        }
 
}
 
 
 
 
 
//Uniform Number Generator
float unifRand()
{
    return rand() /float(RAND_MAX);
}
 
 
int main(){
 
        float *x,*y;
        float *x_dev,*y_dev;
        float p_cpu;
 
 
        float *c_dev;
        float c=0.0;
        int *c_devI;
        int cI=0;
 
        //Allocate mmeory on Host
        x=(float*)malloc(N*sizeof(float));
        y=(float*) malloc(N*sizeof(float));
 
        //Inizialize Host
        for (int i=0;i<N;i++){
                x[i]=unifRand();
                y[i]=unifRand();
        }
 
        //Allocate memory on Device
        cudaMalloc((void**)&x_dev,N*sizeof(float));
        cudaMalloc((void**)&y_dev,N*sizeof(float));
        cudaMalloc((void**)&c_dev,sizeof(float));
        cudaMalloc((void**)&c_devI,sizeof(int));
 
 
 
 
        //Transfer memory from  host to device
        cudaMemcpy(x_dev,x,N*sizeof(float),cudaMemcpyHostToDevice);
        cudaMemcpy(y_dev,y,N*sizeof(float),cudaMemcpyHostToDevice);
        cudaMemcpy(c_dev,&c,sizeof(float),cudaMemcpyHostToDevice);
        cudaMemcpy(c_devI,&cI,sizeof(int),cudaMemcpyHostToDevice);
 
 
 
        p_cpu=piEstimateCPU(x,y);
        printf("CPU pi=%f\n", p_cpu);
 
 
        //Define grid dimension and block dimension
        dim3 mblockDim(THREADS_PER_BLOCK);
        dim3 mgridDim(N/THREADS_PER_BLOCK+1);
        //Call kernel
        piEstimateGPU<<<mgridDim,mblockDim>>>(x_dev,y_dev,c_dev);
        float somma;
        cudaMemcpy(&somma,c_dev,sizeof(float),cudaMemcpyDeviceToHost);
        printf("Pi con atomicAdd %f\n", somma);
 
        //Call kernel
        piEstimateGPUInt<<<mgridDim,mblockDim>>>(x_dev,y_dev,c_devI);
        cudaMemcpy(&cI,c_devI,sizeof(float),cudaMemcpyDeviceToHost);
        somma=4.0*cI/N;
        printf("Pi con atomicAdd %f\n", somma);
 
        return 0;
}
