#include<stdio.h>
#include<stdlib.h>
#include<math.h>


#define N 20480
#define THREADS_PER_BLOCK 256

//CPU Approximation 
float piEstimateCPU(float *x, float *y)
{
    int s=0;
    for( int i = 0; i < N; i++)
    {
    	float xx = x[i];
		float yy = y[i];
		float norm = xx*xx + yy*yy;
		if (norm < 1.0f)
			s++;
    }
    return 4.0*s/N;
}



__global__ void piEstimateGPU(float *x_dev,float *y_dev, float *p_dev){
	
	int idx=blockIdx.x*blockDim.x+threadIdx.x;
	__shared__ int temp[THREADS_PER_BLOCK];	
	
	if(idx>= N) return;
	
	float xx=x_dev[idx];
	float yy=y_dev[idx];
	float norm =xx*xx+yy*yy;
	if (norm <1.0f)  temp[threadIdx.x]=1;
	else temp[threadIdx.x]=0;
	__syncthreads();
	
	if(threadIdx.x==0){	
		int count=0;
		for(int i=0;i<THREADS_PER_BLOCK;i++) count+=temp[i];
		float pi=4.0*count/N;
		atomicAdd(p_dev,pi);
	}
		
	
}
//Uniform Number Generator
float unifRand()
{
    return rand() /float(RAND_MAX);
}


int main(){
	
	float *x,*y;
	float *x_dev,*y_dev;
	float p=0.0, *p_dev;
	float p_cpu;
	//Allocate mmeory on Host
	cudaMallocHost((void**)&x,N*sizeof(float));	
	cudaMallocHost((void**)&y,N*sizeof(float));
	//Inizialize Host  
	for (int i=0;i<N;i++){
		x[i]=unifRand();
		y[i]=unifRand();	
	}
	
	//Allocate memory on Device
	cudaMalloc((void**)&x_dev,N*sizeof(float));	
	cudaMalloc((void**)&y_dev,N*sizeof(float));
	cudaMalloc((void**)&p_dev,sizeof(float));
	//Transfer memory from  host to device
	cudaMemcpy(x_dev,x,N*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(y_dev,y,N*sizeof(float),cudaMemcpyHostToDevice);
	cudaMemcpy(p_dev,&p,sizeof(float),cudaMemcpyHostToDevice);
	
	p_cpu=piEstimateCPU(x, y);
	printf("CPU pi=%f\n", p_cpu);
	
	
	//Define grid dimension and block dimension
	dim3 mblockDim(THREADS_PER_BLOCK);
	dim3 mgridDim(N/THREADS_PER_BLOCK+1);
	//Call kernel
	piEstimateGPU<<<mgridDim,mblockDim>>>(x_dev,y_dev,p_dev);
	cudaMemcpy(&p,p_dev,sizeof(float),cudaMemcpyDeviceToHost);
	printf("GPU pi=%f\n", p);
		
	return 0;
}
